// rearranged LD3320 sample code, for tests on both asr and mp3 playing by ken woo 20220614.v.0.2.
// 20220618 v.0.3, support esp-m3 which disabled tx/rx. and with some structural modifies.
// 20220624 v.0.4, example, now uses station mode and auto connect and supports multiple files upload. add mp3 cont mode.
// 20220626 v.0.5. debug.
// 20220629 v.0.6, LD3320.h. finalized:
// 1. the mp3 files names and sentences are only supported in lower case.
// 2. if no mp3 files or sentences(simply said "mp3s") are playing then asr is running instead.
// 3. if the voice MP3Sentence("how are you today") is trimmed, then use delay instead; MP3Sentence("how are you today", 3).
// 4. by users; it is better to remove all the LittleFS.begin() and .end() and add only one LittleFS.begin() in global.
// 5. littlefs has 8192 bytes of a block. the collection of "word_new" only takes < 1MB however, hereby consumed 3MB in littlefs.
//    it is better to have ways to change 8192 smaller; as such, ESP-M3(1MB) or esp8285 could be utilized.
// A. limitations:
// A.1. if !MP3_IN_ISR, then it is possible falling into idle and no mp3s or asr is running.
//          no solution yet; i have not encountered yet at this version(former versions did).
// A.2. if MP3_IN_ISR, then isr may monopolize and then cause wdt-reset if uses long sentences,
//          if so, shortening it or breaking into multiple sentences or disable wdt.
// A.3. the library "tables" is needed.
// A.4. the lib Esp8266Vars2File is optional which corresponds to macro MYDB; please notice about it also uses littlefs.
// A.5. IoTs code is used; the mp3 files could be bulk uploaded, however if failed at some file, redo beginning it through.
// A.6. at least the mp3 files are needed: _bp.mp3(beep), _sl.mp3(silence).
// A.7. c_strings must static or global otherwise might unable to referenced.
// A.8. if no some necessary IRAM_ATTRs, mp3 play would lag. for other platforms, should rewrite it as the same spdup effects.
// A.9. mp3 files uses 32kbps is fine, higher causes lag or refine transfer rate or procedure might overcome.
// A.10. if some play sound is trimmed, should add trailing silence delay, see contexts.
// A.11. the procedural delays use far smaller than original. tune it if ld3320 not works.
// A.13. mass use print and littlefs caused exists() fault, i do not know why.
// A.14. at least now both MP3_IN_ISR(if wdt disabled or short sentences) and !MP3_IN_ISR are played ok on esp8266.
// A.15. in main-loop while using {LD3320::StopASR(); nAsrStatus=LD_ASR_NONE;}, make sure mp3 is not playing or intr lost/halt.
// A.16. since mp3 play is asynchronous behavior, and asr/mp3 in one chip, so there are many potential troubles to handle.
// A.17. these files are necessary:
//       _sl.mp3, _bp.mp3, _doo.mp3, _error1.mp3, _error2.mp3, _intro1.mp3, _switch1.mp3, startup.mp3, sample?.mp3.
// B. usages: refer to the examples.
// C. todo: none aware.
// D. hopetodone: smaller block of esp8266::littlefs.
//
// 20220629 v.0.6-1, LD3320.h.
//
// 20220704 v.0.7, LD3320.h. add symbol expansion in literal(). known bug: !MP3_IN_ISR occasionally stop ASRing.




#ifndef __LD3320_H
#define __LD3320_H
#include <ESP8266WiFi.h>
#include<LittleFS.h>


////////AAAAAAAAA/////////// user modify here.

#define MP3_IN_ISR      0 // if in isr, the isr has to access flash then may cause problem,
                          // instead, only set flag for later main loop to access, however might affect played performance.
#define ESP8266_SPEEDUP 1 // use esp8266 faster io.

#define USE_ESP_M3      0

#define InTheIram       IRAM_ATTR

#define AUDIO_VOL       7 // 7 is mid. from 0 to 15.

#define LITTLEFS_BUFFER_SIZE 2048 // the number of bytes of the buffer for read mp3 files.

#define MP3_SENTENCE_MAX_LEN 255 // the max size of how many characters/incl.space and even "extra delay" are there in a sentence.

#define MP3_FILEWORD_MAX_LEN 16 // the max size of how many characters are there in a word.
// the max size of how many characters/incl. the begin "_" and the tail ".mp3" are there in an mp3 filename.
#define MP3_FILENAME_MAX_LEN (MP3_FILEWORD_MAX_LEN+5)

#define MP3_DEFAULT_SOUND ("_bp.mp3") // if file not found, use this instead. set either 0 or "somefile.mp3".

#define MP3_SILENCE_SOUND (" sl")     // " sl", the white space is necessary. and the "_sl.mp3" i attached takes 830ms.
                                      // note, exactly only the "_sl.mp3" allowed or use customized with the same format.
                                      // in order to recover LD3320 shortage:
                                      // the use of the dac-out-end flag, LD_ReadReg(0xBA), might later cause ld3320 exception.
                                      // however without use, would cause voice had not been finished but have had been reset.
                                      // so, need include this silence file, and manually added this delay if required.

#if !(USE_ESP_M3)
// modified for esp8266, wemos d1 mini.
    #define SDCK            5 // SPI时钟信号, D1.
    #define SDO             4 // SPI slave data output, MI, 数据输出, D2.
    #define SDI             14 // SPI, MO, 数据输入, D5.
    #define SCS             12 // 芯片片选信号, D6.
    #define RSTB            2 // 复位端口, D4.
    #define IRQ             13 // 中断引脚, D7.
#else // USE_ESP_M3
// modified for esp8285, esp-m3-1m. should use #define MYDEBUG 0 in main().
    #define SDCK            12 // SPI时钟信号
    #define SDO             3 // SPI slave data output, MI, 数据输出
    #define SDI             1 // SPI, MO, 数据输入
    #define SCS             0 // 芯片片选信号
    #define RSTB            14 // 复位端口
    #define IRQ             4 // 中断引脚
#endif

#define MYDELAY1        delayMicroseconds(1)//delay(1)
#define MYDELAY5        delayMicroseconds(10)//delay(5)
#define MYDELAY20       delayMicroseconds(15)//delay(20)
#define MYDELAY_NOP     delayMicroseconds(1)
#define MYDELAY_FIXED1  delay(1)

#if ESP8266_SPEEDUP // for speed up spi timing for esp8266.
    extern volatile unsigned &PIN_OUT_SET;
    extern volatile unsigned &PIN_OUT_CLEAR;
    extern volatile unsigned &PIN_IN;

    #define IO00    0x00001
    #define IO01    0x00002
    #define IO02    0x00004
    #define IO03    0x00008
    #define IO04    0x00010
    #define IO05    0x00020
    #define IO12    0x01000
    #define IO13    0x02000
    #define IO14    0x04000
    #define IO15    0x08000

    #if !(USE_ESP_M3) // wemos d1 mini
        #define SDCK_1 (PIN_OUT_SET=IO05)
        #define SDCK_0 (PIN_OUT_CLEAR=IO05)

        #define SDI_1 (PIN_OUT_SET=IO14)
        #define SDI_0 (PIN_OUT_CLEAR=IO14)

        #define SCS_1 (PIN_OUT_SET=IO12)
        #define SCS_0 (PIN_OUT_CLEAR=IO12)

        #define RSTB_1 (PIN_OUT_SET=IO02)
        #define RSTB_0 (PIN_OUT_CLEAR=IO02)

        #define SDO_READ (PIN_IN&IO04)
    #else // (USE_ESP_M3)
        #define SDCK_1 (PIN_OUT_SET=IO12)
        #define SDCK_0 (PIN_OUT_CLEAR=IO12)

        #define SDI_1 (PIN_OUT_SET=IO01)
        #define SDI_0 (PIN_OUT_CLEAR=IO01)

        #define SCS_1 (PIN_OUT_SET=IO00)
        #define SCS_0 (PIN_OUT_CLEAR=IO00)

        #define RSTB_1 (PIN_OUT_SET=IO14)
        #define RSTB_0 (PIN_OUT_CLEAR=IO14)

        #define SDO_READ (PIN_IN&IO03)
    #endif

#else // ordinary arduino io access.
    #define SDCK_1 digitalWrite(SDCK, HIGH)
    #define SDCK_0 digitalWrite(SDCK, LOW)

    #define SDI_1 digitalWrite(SDI, HIGH)
    #define SDI_0 digitalWrite(SDI, LOW)

    #define SCS_1 digitalWrite(SCS, HIGH)
    #define SCS_0 digitalWrite(SCS, LOW)

    #define RSTB_1 digitalWrite(RSTB, HIGH)
    #define RSTB_0 digitalWrite(RSTB, LOW)

    #define SDO_READ (digitalRead(SDO))
#endif

////////VVVVVVVVV/////////// user modify here.




//  以下三个状态定义用来记录程序是在运行ASR识别还是在运行MP3播放
#define LD_MODE_IDLE      0x00
#define LD_MODE_ASR_RUN   0x08
#define LD_MODE_MP3       0x40


// for playing mp3
#define MASK_FIFO_STATUS_AFULL      0x08 // fifo_data full if bitset.
#define CAUSE_MP3_SONG_END          0x20 // chip saw the mp3-file-end.


// 以下状态定义用来记录程序是在运行ASR识别过程中的哪个状态
#define LD_ASR_IDLE       0x00  //  表示没有在作ASR识别
#define LD_ASR_NONE       0x01  //  available and going do asr or play mp3.
#define LD_ASR_BUSY       0x02  //  表示LD3320正在作ASR识别中
#define LD_ASR_FOUNDOK    0x04  //  表示一次识别流程结束后，有一个识别结果
#define LD_ASR_FOUNDZERO  0x08  //  表示一次识别流程结束后，没有识别结果
#define LD_ASR_ERROR      0x10  //  表示一次识别流程中LD3320芯片内部出现不正确的状态
#define LD_ASR_EXCEPTION  0x11  //  exception
#define LD_ASR_INTERR     0x12  //  interrupt error
#define LD_MP3_INTR       0x14  //  mp3 interrupt
#define LD_MP3_DATA       0x18  //  (for non-isr only) set before mp3 processing data LD_ReloadMp3Data().
#define LD_MP3_REST       0x20  //  reset after MP3 done. for delay purpose.


#define CLK_IN          24  /* 用户注意修改输入的晶振时钟大小 */
#define LD_PLL_11       (unsigned char)((CLK_IN/2.0)-1)
#define LD_PLL_MP3_19   0x0f
#define LD_PLL_MP3_1B   0x18
#define LD_PLL_MP3_1D   (unsigned char)(((90.0*((LD_PLL_11)+1))/(CLK_IN))-1)

#define LD_PLL_ASR_19   (unsigned char)(CLK_IN*32.0/(LD_PLL_11+1) - 0.51)
#define LD_PLL_ASR_1B   0x48
#define LD_PLL_ASR_1D   0x1f


#define MIC_VOL 0x43   //咪头增益


typedef class LD3320{ // all static.

    typedef void (*fptr)();                     // function pointer without parameter.
    typedef void (*p1fptr)(unsigned);           // function pointer with 1 parameter.

    // vars
    InTheIram static unsigned nLD_Mode;         // the mode state, asr/mp3/idle.
    InTheIram static unsigned nLD_reset;        // demand to reset/go for whole procedure, for asr.
    InTheIram static char *(*var_tbl)[];        // the table pointer used for loading the table, for asr.
    InTheIram static unsigned nLD_change_table; // demand to change table, for asr.
    InTheIram static unsigned nLD_IntReg1, nLD_IntReg2; // int registers for backup and restore. while mp3 mode.
    InTheIram static unsigned main_busy;        // to distinguish from main-loop entrance or isr entrance(possible value 0, 1, 2).

    // callback funcs
    IRAM_ATTR static p1fptr ld_fptr;            // for asr. callbacked if set, and used in one-shot. applied when RunASR().
    IRAM_ATTR static void LD_Callback(unsigned id); // for asr. called only if ld_fptr is set, and then set it to 0.
    IRAM_ATTR static unsigned (*ld_mp3_cont_cb)(); // callback function ptr for mp3 continuous mode.
                                                   // it represents the use of continuous mode or not.
                                                   // if used, isr may overwhelm the processes; if so, use !MP3_IN_ISR instead.

    // basics
    IRAM_ATTR static void LD_WriteReg(unsigned address, unsigned dataout);
    IRAM_ATTR static unsigned LD_ReadReg(unsigned address);
    InTheIram static void LD_Init_Common();
    InTheIram static void LD_Init_ASR();
    InTheIram static unsigned LD_AsrRun();
    InTheIram static unsigned LD_AsrAddFixed();
    InTheIram static unsigned LD_Check_ASRBusyFlag_b2();

    // mp3s. note that the play mp3 procedure can not be simplified even successive plays.
    InTheIram static void LD_Init_MP3();
    InTheIram static void LD_Play_MP3();
    IRAM_ATTR static void LD_ReloadMp3Data();
    IRAM_ATTR static void LD_BackupIntsDis();   // for mp3 mode.
    IRAM_ATTR static void LD_RestoreInts();     // for mp3 mode.

        public:

    // basics
    static void LD_Init();
    IRAM_ATTR static void LD_Reset();
    IRAM_ATTR static void LD_INT();
    IRAM_ATTR static unsigned LD_GetResult();

    // asrs
    static unsigned RunASR(p1fptr a=0);
    static void StopASR();

    // mp3s
    #if !(MP3_IN_ISR)
        IRAM_ATTR static void ProcessIntMP3();
    #endif
    static unsigned RunMP3();
    static void LD_AdjustMIX2SPVolume(unsigned val);
    static unsigned EnterMP3ContMode(unsigned (*fptr_cb)()); // if the callback return 0 would leave from this mp3 continuous mode.

    // helper funcs
    static void LD_ChangeTable(void *t); // t, a table pointer.
    static void* LD_GetTable(); // return the current table pointer.
    static void LD_LoopEx(); // main loop example.
    static unsigned LD_loadNextPlayEx(); // mp3 continuous mode callback function example.

} LD3320;

extern unsigned nAsrStatus; // procedural control.




typedef class LD3320MP3{ // note that since it is byte-manipulation, so it can not in IRAM_ATTR(for Read).
    // reasons that, over here is single threaded. only one mp3 file is played at a time.
    // close it before proceeding another file is ok.
    // and the filesystem we use uses buffer for any number of bytes we request which is ok too, i.e., so 1.
    // so we can simply use a static member function to get data and must explicitly close before for another file.
    //
    // !!!note the potential bug: for LittleFS.begin() might not be the ordinary object behavior. and the class MYDB also uses it,
    // so it is better to comment out all the begin() and end() calls and put only one begin() in global.
    // remind that also notice about LD3320MP3("abc") & LD3320MP3(var) & LD3320MP3 t("abc") & LD3320MP3 t(var) might unreferenceable.

    IRAM_ATTR static File fh;
    IRAM_ATTR static unsigned flock;
    InTheIram static const char *var_mp3_filename;
    static unsigned char lfs_buff[LITTLEFS_BUFFER_SIZE]; // can not be IRAM_ATTR.
    IRAM_ATTR static int buff_idx, buff_rsz;

    IRAM_ATTR static int my_read(unsigned char *dst, unsigned n){
        // for simplifying, assuming consume 1 byte each time,
        // however can as many as the current remaining available data in buffer.
        while (1)
            if (buff_rsz){
                if (n>buff_rsz) n=buff_rsz;
                unsigned i=n;
                buff_rsz-=i;
                while (i--) *dst++=lfs_buff[buff_idx++];
                return n;
            }
            else {
                buff_idx=0;
                buff_rsz=fh.read(lfs_buff, LITTLEFS_BUFFER_SIZE);
                if (buff_rsz<=0){
                    buff_rsz=0;
                    return buff_rsz;
                }
                continue;
            }
    };

        public:

    IRAM_ATTR static unsigned Read(unsigned &r_byte){ // since function calls are in flash and inevitable.
                                                      // return true if success, false if file-end or errors.
        int i=0;
        unsigned j;
        if (fh || flock || (fh=LittleFS.open(var_mp3_filename, "r")) ||\
            (MP3_DEFAULT_SOUND && (fh=LittleFS.open(MP3_DEFAULT_SOUND, "r")) && (var_mp3_filename=MP3_DEFAULT_SOUND))\
        ){
            //// if (flock!=2) i=fh.read((unsigned char*)(&j), 1);
            if (flock!=2) i=my_read((unsigned char*)(&j), 1); ////
            if (i!=1){
                if (flock!=2){
                    var_mp3_filename=0;
                    flock=2;
                    fh.close();
                    fh=File(0);
                    LittleFS.end();
                }
                return 0;
            }
            flock=1;
            r_byte=j;
            return 1;
        }
        return 0;
    };

    LD3320MP3(){;};
    LD3320MP3(const char *a){if (var_mp3_filename || !a || !*a) return; LittleFS.begin(); var_mp3_filename=a;};

    IRAM_ATTR static unsigned isDone(){return (flock==2);};
    IRAM_ATTR static unsigned isProc(){return (unsigned)fh;};

    IRAM_ATTR static unsigned isReady(){
        if (var_mp3_filename){
            //////////////// note!!!!!!!!!!!!!!!!!!!!!! exists() might cause wdt-reset//////////////////////////////////////
            if (LittleFS.exists(var_mp3_filename)) return 1;
            else if (MP3_DEFAULT_SOUND && LittleFS.exists(MP3_DEFAULT_SOUND)){var_mp3_filename=MP3_DEFAULT_SOUND; return 1;}

            // the exception. may affect status but might more safe.
            var_mp3_filename=0;
            flock=0;
            if (fh){fh.close(); fh=File(0);}
            LittleFS.end();
        }
        return 0;
    };

    IRAM_ATTR static void Close(){
        if (var_mp3_filename){
            if (fh){fh.close(); fh=File(0);}
            var_mp3_filename=0;
            LittleFS.end();
        }
        flock=0;
    };

    InTheIram static unsigned isOpened(){return (var_mp3_filename || fh || flock);}; // must check closed before playing another file.

} LD3320MP3;




typedef class MP3Sentence{ // e.g., MP3Sentence("how are you"); will sequentially call LD3320MP3("_how.mp3", "_are.mp3", "_you.mp3");
    static char sentence[MP3_SENTENCE_MAX_LEN]; // non null-terminated.
    static char filename[MP3_FILENAME_MAX_LEN+1]; // null-terminated, so plus one.
    static const char extname[5]; // exact the ".mp3". // null-terminated.
    IRAM_ATTR static unsigned iter, endter; // the iterator and the end point.

        public:

    MP3Sentence(){;};

    MP3Sentence(const char *a, unsigned delay_unit=0){ // if 1 unit delay delays 3 seconds. e.g., delay 9 seconds uses 3.
        unsigned i;
        if (iter<endter || !*a) return; // busy or null-string.
        i=0;
        iter=0;
        do sentence[iter]=a[i++]; while (a[i] && ++iter<MP3_SENTENCE_MAX_LEN);
        endter=i;
        iter=0;

        for (i=0; delay_unit--; i=0){ // to add extra silence strings, e.g., " sl" or " sl sl" or " sl sl sl", etc.
            while (MP3_SILENCE_SOUND[i] && endter<MP3_SENTENCE_MAX_LEN) sentence[endter++]=MP3_SILENCE_SOUND[i++];
            if (endter>=MP3_SENTENCE_MAX_LEN){ // means space ran out.
                if (MP3_SILENCE_SOUND[i]){ // means not just make, need to refine.
                    for (; MP3_SILENCE_SOUND[i]!=' '; i--, endter--); // roll back.
                }
                break;
            }
        }

        LD3320::EnterMP3ContMode(MP3Sentence::getNextFilename);
    };

    static void Literal(const char *a, unsigned delay_unit=0){ // for literally converter; "abc" to "a b c".
        unsigned i;
        if (iter<endter || !*a) return; // busy or null-string.
        i=0;
        iter=0;

        // here for expand other functional supports, for example, convert "." to "dot", "/" to "slash".
        //////////////////////////////////////////////////////////////////////////////////////////////////////
        #if 0 // normal behavior.
        do {
            sentence[iter++]=a[i++];
            sentence[iter]=' ';
        }
        while (a[i] && ++iter<MP3_SENTENCE_MAX_LEN);
        //// endter=(a[i]? iter: iter+1); // since the last one is space.
        endter=iter; ////
        iter=0;
        #else // has symbol expansion support.
        //////////////////////////////////////////////////////////////////////////////////////////////////////
        #define _N_ts_DOT               "dot" // ".", dot.
        #define _N_ts_SLASH             "slash" // "/", slash.
        #define _N_ts_DASH              "dash" // "-", dash.
        #define _N_ts_HASH              "hash" // "#", hash.
        #define _N_ts_AT                "at" // "@", at.
        #define _N_ts_COLON             "colon" // ":", colon.
        #define _N_ts_UNDERSCORE        "underscore" // "_", underscore.

        #define _N_tr_DOT               3 // ".", dot.
        #define _N_tr_SLASH             5 // "/", slash.
        #define _N_tr_DASH              4 // "-", dash.
        #define _N_tr_HASH              4 // "#", hash.
        #define _N_tr_AT                2 // "@", at.
        #define _N_tr_COLON             5 // ":", colon.
        #define _N_tr_UNDERSCORE        10 // "_", underscore.

        const char *p; unsigned q;

        while (a[i]){
            if ((a[i]<'a' || a[i]>'z') && (a[i]<'0' || a[i]>'9')){ // if symbols.
                switch (a[i]){
                    case '.':
                        p=_N_ts_DOT;
                        q=_N_tr_DOT;
                        break;
                    case '/':
                        p=_N_ts_SLASH;
                        q=_N_tr_SLASH;
                        break;
                    case '-':
                        p=_N_ts_DASH;
                        q=_N_tr_DASH;
                        break;
                    case '#':
                        p=_N_ts_HASH;
                        q=_N_tr_HASH;
                        break;
                    case '@':
                        p=_N_ts_AT;
                        q=_N_tr_AT;
                        break;
                    case ':':
                        p=_N_ts_COLON;
                        q=_N_tr_COLON;
                        break;
                    case '_':
                        p=_N_ts_UNDERSCORE;
                        q=_N_tr_UNDERSCORE;
                        break;
                    default:
                        p=0;
                        q=0;
                        break;
                }
                if (q){
                    if (iter+q<=MP3_SENTENCE_MAX_LEN) while (q--) sentence[iter++]=*p++;
                    /*else { // means content full, unable succeeding add.
                        endter=iter;
                        iter=0;

                        LD3320::EnterMP3ContMode(MP3Sentence::getNextFilename);
                        return;
                    }*/
                }
                else if (iter<MP3_SENTENCE_MAX_LEN) sentence[iter++]=a[i]; // other symbols.
                /*else { // means content full, unable succeeding add.
                    endter=iter;
                    iter=0;

                    LD3320::EnterMP3ContMode(MP3Sentence::getNextFilename);
                    return;
                }*/
            }
            else if (iter<MP3_SENTENCE_MAX_LEN) sentence[iter++]=a[i]; // ordinary literals.
            /*else { // means content full, unable succeeding add.
                endter=iter;
                iter=0;

                LD3320::EnterMP3ContMode(MP3Sentence::getNextFilename);
                return;
            }*/

            // this line is the code black hole such that all the above else-return could be eliminated and use below else.
            if (iter<MP3_SENTENCE_MAX_LEN) sentence[iter++]=' ';

            else { // means content full, unable succeeding add.
                endter=iter;
                iter=0;

                LD3320::EnterMP3ContMode(MP3Sentence::getNextFilename);
                return;
            }
            i++;
        }
        endter=iter;
        iter=0;
        #endif // choose normal or additional symbol expansion.
        //////////////////////////////////////////////////////////////////////////////////////////////////////

        for (i=0; delay_unit--; i=0){ // to add extra silence strings, e.g., " sl" or " sl sl" or " sl sl sl", etc.
            while (MP3_SILENCE_SOUND[i] && endter<MP3_SENTENCE_MAX_LEN) sentence[endter++]=MP3_SILENCE_SOUND[i++];
            if (endter>=MP3_SENTENCE_MAX_LEN){ // means space ran out.
                if (MP3_SILENCE_SOUND[i]){ // means not just make, need to refine.
                    for (; MP3_SILENCE_SOUND[i]!=' '; i--, endter--); // roll back.
                }
                break;
            }
        }

        LD3320::EnterMP3ContMode(MP3Sentence::getNextFilename);
    };

    static void AddSilence(const char *a, unsigned delay_unit=0){ // play a single file, e.g., a="ex1.mp3" or multiple files,
                                                                  // e.g., a="ex1.mp3 ex2.mp3" with trailing silence files.
                                                                  // the source string of the silence file name is supposed to be
                                                                  // " filename" which will be reworked to become "_filename.mp3",
                                                                  // finally is e.g., "ex1.mp3 ex2.mp3 _filename.mp3 _filename.mp3".
        unsigned i;
        if (iter<endter || !*a) return; // busy or null-string.
        i=0;
        iter=0;
        do sentence[iter]=a[i++]; while (a[i] && ++iter<MP3_SENTENCE_MAX_LEN);
        endter=i;
        iter=0;

        strcpy(filename+1, MP3_SILENCE_SOUND+1);
        strcpy(filename+strlen(filename), extname);
        i=strlen(filename)+1; // with a space.

        // to add extra silence strings, e.g., "_sl.mp3" or "_sl.mp3 _sl.mp3", etc.
        for (; delay_unit-- && ((endter+i)<=MP3_SENTENCE_MAX_LEN); endter+=i){
            sentence[endter]=' ';
            strcpy(sentence+endter+1, filename);
        }

        LD3320::EnterMP3ContMode(MP3Sentence::getNextFilenameAlt);
    };

    static unsigned isIdle(){return iter>=endter;}; // need to wait for idle to send sentence.

    InTheIram static unsigned getNextFilename(){ // the purpose is to append/extract every word, e.g., "to" becomes "_to.mp3".
        // need IRAM_ATTR since operational intense however might fault since byte 'a'.
        // if so, add such as (unsigned)'a'.
        unsigned i=1; // for filename[0]='_';
        unsigned onechance=1;
        unsigned lock=iter;

        while (lock<endter){
            // non-alphabet or non-numerical. note that we only support lower case latters.
            if (((sentence[lock]<'a' || sentence[lock]>'z') && (sentence[lock]<'0' || sentence[lock]>'9'))){
                lock++;
                if (onechance) continue;
                else {
                    iter=lock;
                    strcpy(filename+i, extname); // to fulfill the remaining part ".mp3".
                    while (LD3320MP3::isOpened() && !LD3320MP3::isDone()) delay(1);             // <4 lately closing>
                    LD3320MP3::Close();                                                         // <4 lately closing>
                    LD3320MP3 t(filename);
                    return 1; // note that LD3320MP3(filename) did not work. ////////////////////////////////////////
                }
            }
            if (onechance) onechance=0;
            if (i>MP3_FILEWORD_MAX_LEN){ // a word overflow.
                filename[1]=0;
                iter=lock;
                return 0; //////////////////////////////// no close, isr would do that //////////////////////////////
            }
            filename[i++]=sentence[lock++];
        }
        iter=lock;
        if (i==1) return 0; /////////////////////////////////////////////////////////////////////////////////////////
        strcpy(filename+i, extname); // to fulfill the remaining part ".mp3".
        while (LD3320MP3::isOpened() && !LD3320MP3::isDone()) delay(1);                         // <4 lately closing>
        LD3320MP3::Close();                                                                     // <4 lately closing>
        LD3320MP3 t(filename);
        return 1; // note that LD3320MP3(filename) did not work. ////////////////////////////////////////////////////
    };

    InTheIram static unsigned getNextFilenameAlt(){ // the diff from getNextFilename() is only extract no rewrite.
        // need IRAM_ATTR since operational intense however might fault since byte 'a'.
        // if so, add such as (unsigned)'a'.
        unsigned i=1; // for filename[0]='_';
        unsigned onechance=1;
        unsigned lock=iter;

        while (lock<endter){
            // (non-alphabet or non-numerical) and not (. or _). note that we only support lower case latters.
            if (
                ((sentence[lock]<'a' || sentence[lock]>'z') && (sentence[lock]<'0' || sentence[lock]>'9')) &&\
                (sentence[lock]!='.') &&\
                (sentence[lock]!='_')\
            ){
                lock++;
                if (onechance) continue;
                else {
                    iter=lock;
                    filename[i]=0;
                    while (LD3320MP3::isOpened() && !LD3320MP3::isDone()) delay(1);             // <4 lately closing>
                    LD3320MP3::Close();                                                         // <4 lately closing>
                    LD3320MP3 t(filename+1);
                    return 1; // note that LD3320MP3(filename) did not work. ////////////////////////////////////////
                }
            }
            if (onechance) onechance=0;
            if (i>MP3_FILEWORD_MAX_LEN){ // a word overflow.
                filename[1]=0;
                iter=lock;
                return 0; ///////////////////////////////////////////////////////////////////////////////////////////
            }
            filename[i++]=sentence[lock++];
        }
        iter=lock;
        if (i==1) return 0; /////////////////////////////////////////////////////////////////////////////////////////
        filename[i]=0;
        while (LD3320MP3::isOpened() && !LD3320MP3::isDone()) delay(1);                         // <4 lately closing>
        LD3320MP3::Close();                                                                     // <4 lately closing>
        LD3320MP3 t(filename+1);
        return 1; // note that LD3320MP3(filename) did not work. ////////////////////////////////////////////////////
    };

} MP3Sentence;


// -------- start/stop ld3320.
inline void LD3320Start(){
    LD3320MP3::Close();
    nAsrStatus=LD_ASR_IDLE;
    LD3320::LD_Init(); // to determined state.
    LD3320::LD_Reset();
    nAsrStatus=LD_ASR_NONE;
}

inline void LD3320Stop(){
    LD3320MP3::Close();
    nAsrStatus=LD_ASR_IDLE;
    LD3320::LD_Init(); // to determined state.
    LD3320::LD_Reset();
    nAsrStatus=LD_ASR_IDLE;
}


#endif // __LD3320_H
